#------------------------------------------------------------------------------
#
#   listing.py
#
#   Package module file for the SB-Assembler sbasm
#   See www.sbprojects.com for details
#
#   This file contains all the listing functions.
#
#------------------------------------------------------------------------------

import sys, os
import dec

#------------------------------------------------------------------------------

def ListLine (lineno, line):

    """
    Only in Pass 2, and only if listing is allowed according to the settings
    of the list flags and current condition, list the entire line.
    This includes first address, all generated bytes (if listed), IMS prefixes,
    line number, instruction time and source line.
    """

    global Asm, Flags
    
    if dec.Asm.Pass == 2 and ListFlags():
        if not dec.Flags.ErrorInLine:
            if line == chr(26) + " ":
                # Don't list lines containing only Ctrl-Z
                return
            if dec.Asm.Mnemonic[0:3] == ".LF" or dec.Asm.Mnemonic[0:3] == ".LI":
                # Don't list lines containing .LF or .LI directives
                return
            if dec.Asm.List_Line == "":
                dec.Asm.List_Line = ListAddress()
            count = dec.Asm.List_Address
            if count == dec.Asm.BOL_Address:
                dec.Asm.List_Line = dec.Asm.List_Line + ' ' * (4*3)
                if dec.Asm.Mnemonic in ('.EQ', '.SE'):
                    # Avoid jagged alignment if value > $FFFF for declarations
                    dec.Asm.List_Line = dec.Asm.List_Line[:5 + 4 * 3]
            else:
                dec.Asm.List_Line = dec.Asm.List_Line + " " * (4 - dec.Asm.List_Byte_Cnt) * 3
            timing = ListTiming()
            dec.Asm.List_Line = dec.Asm.List_Line + ListLineNo(lineno) + timing +  line.rstrip()
            DoList()
    dec.Flags.ErrorInLine = False

#-----------------------------------------------------------------------------

def ListTiming():

    timing = ''

    if dec.Flags.Tlist and dec.Asm.Timing_Length > 0:
        timing = (' ' * dec.Asm.Timing_Length + dec.Asm.Timing)[-dec.Asm.Timing_Length:]
        if timing.isspace():
            timing = timing + '   '
        else:
            timing = '(' + timing + ') '
    
    return timing

#-----------------------------------------------------------------------------

def DoList():

    """
    Write the current list line to the currently open list file. And then
    erase the list line.
    """
    
    global Asm

    parts = dec.Asm.List_Line.rstrip().split(dec.EOL)
    message = ""
    for i in parts:
        message = message + i[:dec.Asm.List_Length] + dec.EOL
    
    dec.Asm.List_File.write(message)
    dec.Asm.List_Line = ""

#-----------------------------------------------------------------------------

def ListLineNo (lineno):

    """
    Return the formatted current line number to the caller.
    Two prefixes can be given: An I if the line came from an include file
    And an M (line came from macro) or an S (line was skipped because of a
    false conditional assembly).
    The line number will have a width of 5 characters, unless the line number
    is larger than 5 digits.    print address
    """

    prefix1 = " "
    if len(dec.Asm.File_Name) > 1:
        if dec.Asm.File_LineNo[1] != 0:     # Don't for main .IN line yet!
            prefix1 = "I"
    prefix2 = " "
    if len(dec.Asm.Macro_Stack) > 0:
        if dec.Asm.Macro_Stack[0][1] > 0:
            prefix2 = "M"
    if dec.Asm.Cond_False > 0:
        prefix2 = "S"
    return prefix1 + prefix2 + str(lineno).rjust(5) + ' '

#-----------------------------------------------------------------------------

def ListAddress():

    """
    Print the current list address in hex notation.
    Print only 4 digits if MSB 4 digits are all 0.
    Print only 6 digits if MSB 2 digits are all 0.
    Otherwise print all 8 digits.
    """

    decadd = dec.Asm.List_Address * 1L
    if decadd < 0:
        decadd = decadd & dec.MAX32
    address = '0' * 7 + hex(decadd)[2:-1]
    address = address[-8:] + '-'
    if address[:4] == '0000':
        address = address[4:]
    elif address[:2] == '00':
        address = address[2:]
    return address.upper()

#-----------------------------------------------------------------------------

def ListByte(byte):

    global Asm
    
    if dec.Asm.Pass == 2:
        if dec.Asm.List_Line == "":
            dec.Asm.List_Line = ListAddress()
            dec.Asm.List_Byte_Cnt = 0
        else:
            if dec.Asm.List_Byte_Cnt == 4:
                dec.Asm.List_Line = dec.Asm.List_Line + dec.EOL + ListAddress()
                dec.Asm.List_Byte_Cnt = 0
        hexbyte = '0' + hex(byte)[2:]
        if hexbyte[-1] == "L":
            hexbyte = hexbyte[:-1]
        hexbyte = hexbyte[-2:].upper()
        dec.Asm.List_Line = dec.Asm.List_Line + hexbyte + ' '
        dec.Asm.List_Byte_Cnt = dec.Asm.List_Byte_Cnt + 1
 
#-----------------------------------------------------------------------------

def PrintErrors (errors, warnings):

    """
    Print the final number of errors and warnings at the end of the list file,
    if one is open that is.
    """
    
    if dec.Asm.List_File != sys.stdout:
        dec.Asm.List_File.write(errors)
        dec.Asm.List_File.write(warnings)
 
#-----------------------------------------------------------------------------

def ListFlags():

    """
    Combine all list flags to see whether we can list this line or not.
    """
    
    if not dec.Flags.List:
        # Global list flag is off, don't list anything
        return False
    if (len(dec.Asm.Macro_Stack) > 0) and (not dec.Flags.Mlist):
        # We're in a macro, and the macro flag is off
        if dec.Asm.Macro_Stack[0][1] > 0:
            # Ignoring first macro call line, the line containing the first >
            return False
    if (dec.Asm.Cond_False > 0) and (not dec.Flags.Clist):
        # Condition is false and the conditional list flag is not set
        return False
    # We've passed all tests, list this line
    return True
 
#-----------------------------------------------------------------------------

if __name__ == '__main__':
    print
    print "This is a python module, it's not a program."
    print "This module is part of the sbasm package."
    print "Run sbasm instead."
    print

